# -*- coding: utf-8 -*-
from abc import ABC
from copy import deepcopy
import numpy as np
import warnings

from ..state import BranchSupplemental, State
from .move import Move


__all__ = ["GroupMove"]


class GroupMove(Move, ABC):
    """
    A "group" ensemble move based on the :class:`eryn.moves.RedBlueMove`.

    In moves like the :class:`eryn.moves.StretchMove`, the complimentary
    group for which the proposal is used is chosen from the current points in
    the ensemble. In "group" moves the complimentary group is a stationary group
    that is updated every `n_iter_update` iterations. This update is performed with the
    last set of coordinates to maintain detailed balance.

    Args:
        nfriends (int, optional): The number of friends to draw from as the complimentary
            ensemble. This group is determined from the stationary group. If ``None``, it will
            be set to the number of walkers. (default: ``None``)
        n_iter_update (int, optional): Number of iterations to run before updating the
            stationary distribution. (default: 100).
        live_dangerously (bool, optional): If ``True``, allow for ``n_iter_update == 1``.
            (deafault: ``False``)

    ``kwargs`` are passed to :class:`Move` class.

    """

    def __init__(
        self, nfriends=None, n_iter_update=100, live_dangerously=False, **kwargs
    ):

        Move.__init__(self, **kwargs)
        self.nfriends = int(nfriends)
        self.n_iter_update = n_iter_update

        if self.n_iter_update <= 1 and not live_dangerously:
            raise ValueError("n_iter_update must be greather than or equal to 2.")

        self.iter = 0

    def find_friends(self, name, s, s_inds=None, branch_supps=None):
        """Function for finding friends.

        Args:
            name (str): Branch name for proposal coordinates.
            s (np.ndarray): Coordinates array for the points to be moved.
            s_inds (np.ndarray, optional): ``inds`` arrays that represent which leaves are present.
                (default: ``None``)
            branch_supps (dict, optional): Keys are ``branch_names`` and values are
                :class:`BranchSupplemental` objects. For group proposals,
                ``branch_supps`` are the best device for passing and tracking useful
                information. (default: ``None``)

        Return:
            np.ndarray: Complimentary values.

        """
        raise NotImplementedError

    def choose_c_vals(self, name, s, s_inds=None, branch_supps=None):
        """Get the complimentary values."""
        return self.find_friends(name, s, s_inds=s_inds, branch_supps=branch_supps)

    def setup(self, branches):
        """Any setup necessary for the proposal"""
        pass

    def setup_friends(self, branches):
        """Setup anything for finding friends.

        Args:
            branches (dict): Dictionary with all the current branches in the sampler.

        """
        raise NotImplementedError

    def fix_friends(self, branches):
        """Fix any friends that were born through RJ.

        This function is not required. If not implemented, it will just return immediately.

        Args:
            branches (dict): Dictionary with all the current branches in the sampler.

        """
        return

    @classmethod
    def get_proposal(self, s_all, random, gibbs_ndim=None, s_inds_all=None, **kwargs):
        """Generate group stretch proposal coordinates

        Args:
            s_all (dict): Keys are ``branch_names`` and values are coordinates
                for which a proposal is to be generated.
            random (object): Random state object.
            gibbs_ndim (int or np.ndarray, optional): If Gibbs sampling, this indicates
                the true dimension. If given as an array, must have shape ``(ntemps, nwalkers)``.
                See the tutorial for more information.
                (default: ``None``)
            s_inds_all (dict, optional): Keys are ``branch_names`` and values are
                ``inds`` arrays indicating which leaves are currently used. (default: ``None``)

        Returns:
            tuple: First entry is new positions. Second entry is detailed balance factors.

        Raises:
            ValueError: Issues with dimensionality.

        """

        raise NotImplementedError("The proposal must be implemented by " "subclasses")

    def propose(self, model, state):
        """Use the move to generate a proposal and compute the acceptance

        Args:
            model (:class:`eryn.model.Model`): Carrier of sampler information.
            state (:class:`State`): Current state of the sampler.

        Returns:
            tuple: (state, accepted)
                The first return is the state of the sampler after the move.
                The second return value is the accepted count array.

        """

        # Check that the dimensions are compatible.
        ndim_total = 0
        for branch in state.branches.values():
            ntemps, nwalkers, nleaves_, ndim_ = branch.shape
            ndim_total += ndim_ * nleaves_

        if self.nfriends is None:
            self.nfriends = nwalkers

        # Run any move-specific setup.
        self.setup(state.branches)

        if self.iter == 0 or self.iter % self.n_iter_update == 0:
            self.setup_friends(state.branches)

        if self.iter != 0 and self.iter % self.n_iter_update == 0:
            # store old values to maintain detailed balance when updating
            old_branches = deepcopy(state.branches)

        # fix any friends that may have come through rj
        if self.iter != 0 and self.iter % self.n_iter_update != 0:
            self.fix_friends(state.branches)

        # Split the ensemble in half and iterate over these two halves.
        accepted = np.zeros((ntemps, nwalkers), dtype=bool)

        all_branch_names = list(state.branches.keys())

        # get gibbs sampling information
        for branch_names_run, inds_run in self.gibbs_sampling_setup_iterator(
            all_branch_names
        ):

            if not np.all(
                np.asarray(list(state.branches_supplemental.values())) == None
            ):
                new_branch_supps = deepcopy(state.branches_supplemental)
            else:
                new_branch_supps = None

            if state.supplemental is not None:
                new_supps = deepcopy(state.supplemental)
            else:
                new_supps = None

            # setup proposals based on Gibbs sampling
            (
                coords_going_for_proposal,
                inds_going_for_proposal,
                at_least_one_proposal,
            ) = self.setup_proposals(
                branch_names_run, inds_run, state.branches_coords, state.branches_inds
            )

            if not at_least_one_proposal:
                continue

            # need to trick stretch proposal into using the dimenionality associated
            # with Gibbs sampling if it is being used
            gibbs_ndim = 0
            for brn, ir in zip(branch_names_run, inds_run):
                if ir is not None:
                    gibbs_ndim += ir.sum()
                else:
                    # nleaves * ndim
                    gibbs_ndim += np.prod(state.branches[brn].shape[-2:])

            self.current_model = model
            self.current_state = state
            # Get the move-specific proposal.
            q, factors = self.get_proposal(
                coords_going_for_proposal,
                model.random,
                gibbs_ndim=gibbs_ndim,
                s_inds_all=inds_going_for_proposal,
                branch_supps=new_branch_supps,
            )

            # account for gibbs sampling
            self.cleanup_proposals_gibbs(
                branch_names_run, inds_run, q, state.branches_coords
            )

            # order everything properly
            q, _, new_branch_supps = self.ensure_ordering(
                list(state.branches.keys()), q, state.branches_inds, new_branch_supps
            )

            # Compute prior of the proposed position
            # new_inds_prior is adjusted if product-space is used
            logp = model.compute_log_prior_fn(q, inds=state.branches_inds)

            self.fix_logp_gibbs(branch_names_run, inds_run, logp, state.branches_inds)

            # Can adjust supplementals in place
            logl, new_blobs = model.compute_log_like_fn(
                q,
                inds=state.branches_inds,
                logp=logp,
                supps=new_supps,
                branch_supps=new_branch_supps,
            )

            # get log posterior
            logP = self.compute_log_posterior(logl, logp)

            # get previous information
            prev_logl = state.log_like

            prev_logp = state.log_prior

            # takes care of tempering
            prev_logP = self.compute_log_posterior(prev_logl, prev_logp)

            # difference
            lnpdiff = factors + logP - prev_logP

            # draw against acceptance fraction
            accepted = lnpdiff > np.log(model.random.rand(ntemps, nwalkers))

            # Update the parameters
            new_state = State(
                q,
                log_like=logl,
                log_prior=logp,
                blobs=new_blobs,
                inds=state.branches_inds,
                supplemental=new_supps,
                branch_supplemental=new_branch_supps,
            )
            state = self.update(state, new_state, accepted)

            # add to move-specific accepted information
            self.accepted += accepted
            self.num_proposals += 1

        if self.temperature_control is not None:
            state = self.temperature_control.temper_comps(state)

        if self.iter != 0 and self.iter % self.n_iter_update == 0:
            # use old values to maintain detailed balance when updating
            # nfriends
            self.setup_friends(old_branches)

        self.iter += 1
        return state, accepted



# -*- coding: utf-8 -*-

import numpy as np
from copy import deepcopy
from ..state import State
from .move import Move

__all__ = ["MHMove"]


class MHMove(Move):
    r"""A general Metropolis-Hastings proposal

    Concrete implementations can be made by providing a ``get_proposal`` method.
    For standard Gaussian Metropolis moves, :class:`moves.GaussianMove` can be used.

    """

    def __init__(self, **kwargs):

        Move.__init__(self, **kwargs)

    def get_proposal(self, branches_coords, random, branches_inds=None, **kwargs):
        """Get proposal

        Args:
            branches_coords (dict): Keys are ``branch_names`` and values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim] representing
                coordinates for walkers.
            random (object): Current random state object.
            branches_inds (dict, optional): Keys are ``branch_names`` and values are
                np.ndarray[ntemps, nwalkers, nleaves_max] representing which
                leaves are currently being used. (default: ``None``)
            **kwargs (ignored): This is added for compatibility. It is ignored in this function.

        Returns:
            tuple: (Proposed coordinates, factors) -> (dict, np.ndarray)

        Raises:
            NotImplementedError: If proposal is not implemented in a subclass.

        """

        raise NotImplementedError("The proposal must be implemented by " "subclasses")

    def setup(self, branches_coords):
        """Any setup for the proposal.

        Args:
            branches_coords (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim]. These are the curent
                coordinates for all the walkers.

        """

    def propose(self, model, state):
        """Use the move to generate a proposal and compute the acceptance

        Args:
            model (:class:`eryn.model.Model`): Carrier of sampler information.
            state (:class:`State`): Current state of the sampler.

        Returns:
            :class:`State`: State of sampler after proposal is complete.

        """

        self.setup(state.branches_coords)

        # get all branch names for gibbs setup
        all_branch_names = list(state.branches.keys())

        # get initial shape information
        ntemps, nwalkers, _, _ = state.branches[all_branch_names[0]].shape

        # in case there are no leaves yet
        accepted = np.zeros((ntemps, nwalkers), dtype=bool)

        # iterate through gibbs setup
        for branch_names_run, inds_run in self.gibbs_sampling_setup_iterator(
            all_branch_names
        ):
            # setup supplemental information
            if not np.all(
                np.asarray(list(state.branches_supplemental.values())) == None
            ):
                new_branch_supps = deepcopy(state.branches_supplemental)
            else:
                new_branch_supps = None

            if state.supplemental is not None:
                new_supps = deepcopy(state.supplemental)
            else:
                new_supps = None

            # setup information according to gibbs info
            (
                coords_going_for_proposal,
                inds_going_for_proposal,
                at_least_one_proposal,
            ) = self.setup_proposals(
                branch_names_run, inds_run, state.branches_coords, state.branches_inds
            )

            # if no walkers are actually being proposed
            if not at_least_one_proposal:
                continue

            self.current_model = model
            self.current_state = state

            # Get the move-specific proposal.
            q, factors = self.get_proposal(
                coords_going_for_proposal,
                model.random,
                branches_inds=inds_going_for_proposal,
                supps=new_supps,
                branch_supps=new_branch_supps,
            )

            # account for gibbs sampling
            self.cleanup_proposals_gibbs(
                branch_names_run, inds_run, q, state.branches_coords
            )

            # order everything properly
            q, _, new_branch_supps = self.ensure_ordering(
                list(state.branches.keys()), q, state.branches_inds, new_branch_supps
            )

            # if not wrapping with mutliple try (normal route)
            if not hasattr(self, "mt_ll") or not hasattr(self, "mt_lp"):
                # Compute prior of the proposed position
                logp = model.compute_log_prior_fn(q, inds=state.branches_inds)

                self.fix_logp_gibbs(
                    branch_names_run, inds_run, logp, state.branches_inds
                )

                # Compute the lnprobs of the proposed position.
                # Can adjust supplementals in place
                logl, new_blobs = model.compute_log_like_fn(
                    q,
                    inds=state.branches_inds,
                    logp=logp,
                    supps=new_supps,
                    branch_supps=new_branch_supps,
                )

            else:
                # if already computed in multiple try
                logl = self.mt_ll
                logp = self.mt_lp
                new_blobs = None

            # get log posterior
            logP = self.compute_log_posterior(logl, logp)

            # get previous information
            prev_logl = state.log_like

            prev_logp = state.log_prior

            # takes care of tempering
            prev_logP = self.compute_log_posterior(prev_logl, prev_logp)

            # difference
            lnpdiff = factors + logP - prev_logP

            # draw against acceptance fraction
            accepted = lnpdiff > np.log(model.random.rand(ntemps, nwalkers))

            # Update the parameters
            new_state = State(
                q,
                log_like=logl,
                log_prior=logp,
                blobs=new_blobs,
                inds=state.branches_inds,
                supplemental=new_supps,
                branch_supplemental=new_branch_supps,
            )
            state = self.update(state, new_state, accepted)

            # add to move-specific accepted information
            self.accepted += accepted
            self.num_proposals += 1

        # temperature swaps
        if self.temperature_control is not None:
            state = self.temperature_control.temper_comps(state)

        return state, accepted



# -*- coding: utf-8 -*-

from ..state import BranchSupplemental
from . import Move
import numpy as np
import tqdm

__all__ = ["CombineMove"]


class CombineMove(Move):
    """Move that combines specific moves in order.

    Args:
        moves (list): List of moves, similar to how ``moves`` is submitted
            to :class:`eryn.ensemble.EnsembleSampler`. If weights are provided,
            they will be ignored.
        *args (tuple, optional): args to be passed to :class:`Move`.
        verbose (bool, optional): If ``True``, use ``tqdm`` to show progress throught steps.
            This can be very helpful when debugging.
        **kwargs (dict, optional): kwargs to be passed to :class:`Move`.


    """

    def __init__(self, moves, *args, verbose=False, **kwargs):
        # store moves
        self.moves = moves
        self.verbose = verbose
        Move.__init__(self, *args, **kwargs)

    @property
    def accepted(self):
        """Accepted counts for each move."""
        if self._accepted is None:
            raise ValueError(
                "accepted must be inititalized with the init_accepted function if you want to use it."
            )
        # this retrieves the accepted arrays from the individual moves
        accepted_out = [move.accepted for move in self.moves]
        return accepted_out

    @accepted.setter
    def accepted(self, accepted):
        # set the accepted arrays for all moves
        assert isinstance(accepted, np.ndarray)
        for move in self.moves:
            move.accepted = accepted.copy()

    @property
    def acceptance_fraction(self):
        """get acceptance fraction averaged over all moves"""
        acceptance_fraction_out = np.mean(
            [move.acceptance_fraction for move in self.moves], axis=0
        )
        return acceptance_fraction_out

    @property
    def acceptance_fraction_separate(self):
        """get acceptance fraction from each move"""
        acceptance_fraction_out = [move.acceptance_fraction for move in self.moves]
        return acceptance_fraction_out

    @property
    def temperature_control(self):
        """temperature controller"""
        return self._temperature_control

    @temperature_control.setter
    def temperature_control(self, temperature_control):
        # when setting the temperature control object
        # need to apply it to each move
        for i, move in enumerate(self.moves):
            # if weights were provided with moves, remove move class
            if isinstance(move, tuple):
                move = move[0]
            # set temperature control for each move
            move.temperature_control = temperature_control

        # main temperature control here for reference
        self._temperature_control = temperature_control

    @property
    def periodic(self):
        """periodic parameter information"""
        return self._periodic

    @periodic.setter
    def periodic(self, periodic):
        # when setting the periodic parameters
        # need to apply it to each move
        for i, move in enumerate(self.moves):
            # if weights were provided with moves, remove move class
            if isinstance(move, tuple):
                move = move[0]
            move.periodic = periodic
        self._periodic = periodic

    def propose(self, model, state):
        """Propose a combined move.

        Args:
            model (:class:`eryn.model.Model`): Carrier of sampler information.
            state (:class:`State`): Current state of the sampler.

        Returns:
            tuple: (state, accepted)
                The first return is the state of the sampler after the move.
                The second return value is the accepted count array for each walker
                counting for all proposals.

        """

        # prepare for verbosity if needed
        iterator = enumerate(self.moves)
        if self.verbose:
            iterator = tqdm.tqdm(iterator)

        # we will set this inside the loop during the first iteration
        accepted_out = None
        for i, move in iterator:
            # get move out of tuple
            if isinstance(move, tuple):
                move = move[0]

            # run move
            state, accepted = move.propose(model, state)

            # set (first iteration) or add (after first iteration) to accepted_out
            if accepted_out is None:
                accepted_out = accepted.copy()
            else:
                accepted_out += accepted

        return state, accepted_out



from multiprocessing.sharedctypes import Value
import numpy as np
import warnings
from copy import deepcopy
from abc import ABC

# from scipy.special import logsumexp

try:
    import cupy as cp

    gpu_available = True
except (ModuleNotFoundError, ImportError):
    import numpy as cp

    gpu_available = False

from .rj import ReversibleJumpMove
from ..prior import ProbDistContainer
from ..utils.utility import groups_from_inds

___ = ["MultipleTryMove"]


def logsumexp(a, axis=None, xp=None):
    if xp is None:
        xp = np

    max = xp.max(a, axis=axis)
    ds = a - max[:, None]

    sum_of_exp = xp.exp(ds).sum(axis=axis)
    return max + xp.log(sum_of_exp)


def get_mt_computations(logP, log_proposal_pdf, symmetric=False, xp=None):

    if xp is None:
        xp = np

        # set weights based on if symmetric
    if symmetric:
        log_importance_weights = logP
    else:
        log_importance_weights = logP - log_proposal_pdf

    # get the sum of weights
    log_sum_weights = logsumexp(log_importance_weights, axis=-1, xp=xp)

    # probs = wi / sum(wi)
    log_of_probs = log_importance_weights - log_sum_weights[:, None]

    # probabilities to choose try
    probs = xp.exp(log_of_probs)

    # draw based on likelihood
    inds_keep = (probs.cumsum(1) > xp.random.rand(probs.shape[0])[:, None]).argmax(1)

    return log_importance_weights, log_sum_weights, inds_keep


class MultipleTryMove(ABC):
    """Generate multiple proposal tries.

    This class should be inherited by another proposal class
    with the ``@classmethods`` overwritten. See :class:`eryn.moves.MTDistGenMove`
    and :class:`MTDistGenRJ` for examples.

    Args:
        num_try (int, optional): Number of tries. (default: 1)
        independent (bool, optional): Set to ``True`` if the proposal is independent of the current points.
            (default: ``False``).
        symmetric (bool, optional): Set to ``True`` if the proposal is symmetric.
            (default: ``False``).
        rj (bool, optional): Set to ``True`` if this is a nested reversible jump proposal.
            (default: ``False``).
        **kwargs (dict, optional): for compatibility with other proposals.

        Raises:
            ValueError: Input issues.

    """

    def __init__(
        self,
        num_try=1,
        independent=False,
        symmetric=False,
        rj=False,
        use_gpu=None,
        **kwargs
    ):
        self.num_try = num_try

        self.independent = independent
        self.symmetric = symmetric
        self.rj = rj

        if self.rj:
            if self.symmetric or self.independent:
                raise ValueError(
                    "If rj==True, symmetric and independt must both be False."
                )

        self.use_gpu = use_gpu

    @property
    def xp(self):
        xp = cp if self.use_gpu else np
        return xp

    @classmethod
    def special_like_func(self, generated_coords, *args, inds_leaves_rj=None, **kwargs):
        """Calculate the Likelihood for sampled points.

        Args:
            generated_coords (np.ndarray): Generated coordinates with shape ``(number of independent walkers, num_try)``.
            *args (tuple, optional): additional arguments passed by overwriting the
                ``get_proposal`` function and passing ``args_like`` keyword argument.
            inds_leaves_rj (np.ndarray): Index into each individual walker giving the
                leaf index associated with this proposal. Should only be used if ``self.rj is True``. (default: ``None``)
            **kwargs (tuple, optional): additional keyword arguments passed by overwriting the
                ``get_proposal`` function and passing ``kwargs_like`` keyword argument.

        Returns:
            np.ndarray: Likelihood values with shape ``(generated_coords.shape[0], num_try).``

        Raises:
            NotImplementedError: Function not included.

        """
        raise NotImplementedError

    @classmethod
    def special_prior_func(self, generated_coords, *args, **kwargs):
        """Calculate the Prior for sampled points.

        Args:
            generated_coords (np.ndarray): Generated coordinates with shape ``(number of independent walkers, num_try)``.
            *args (tuple, optional): additional arguments passed by overwriting the
                ``get_proposal`` function and passing ``args_prior`` keyword argument.
            inds_leaves_rj (np.ndarray): Index into each individual walker giving the
                leaf index associated with this proposal. Should only be used if ``self.rj is True``. (default: ``None``)
            **kwargs (tuple, optional): additional keyword arguments passed by overwriting the
                ``get_proposal`` function and passing ``kwargs_prior`` keyword argument.

        Returns:
            np.ndarray: Prior values with shape ``(generated_coords.shape[0], num_try).``

        Raises:
            NotImplementedError: Function not included.

        """
        raise NotImplementedError

    @classmethod
    def special_generate_func(
        coords, random, size=1, *args, fill_tuple=None, fill_values=None, **kwargs
    ):
        """Generate samples and calculate the logpdf of their proposal function.

        Args:
            coords (np.ndarray): Current coordinates of walkers.
            random (obj): Random generator.
            *args (tuple, optional): additional arguments passed by overwriting the
                ``get_proposal`` function and passing ``args_generate`` keyword argument.
            size (int, optional): Number of tries to generate.
            fill_tuple (tuple, optional): Length 2 tuple with the indexing of which values to fill
                when generating. Can be used for auxillary proposals or reverse RJ proposals. First index is the index into walkers and the second index is
                the index into the number of tries. (default: ``None``)
            fill_values (np.ndarray): values to fill associated with ``fill_tuple``. Should
                have size ``(len(fill_tuple[0]), ndim)``. (default: ``None``).
            **kwargs (tuple, optional): additional keyword arguments passed by overwriting the
                ``get_proposal`` function and passing ``kwargs_generate`` keyword argument.

        Returns:
            tuple: (generated points, logpdf of generated points).

        Raises:
            NotImplementedError: Function not included.

        """
        raise NotImplementedError

    @classmethod
    def special_generate_logpdf(self, coords):
        """Get logpdf of generated coordinates.

        Args:
            coords (np.ndarray): Current coordinates of walkers.

        Returns:
            np.ndarray: logpdf of generated points.

        Raises:
            NotImplementedError: Function not included.
        """
        raise NotImplementedError

    def get_mt_log_posterior(self, ll, lp, betas=None):
        """Calculate the log of the posterior for all tries.

        Args:
            ll (np.ndarray): Log Likelihood values with shape ``(nwalkers, num_tries)``.
            lp (np.ndarray): Log Prior values with shape ``(nwalkers, num_tries)``.
            betas (np.ndarray, optional): Inverse temperatures to include in log Posterior computation.
                (default: ``None``)

        Returns:
            np.ndarray: Log of the Posterior with shape ``(nwalkers, num_tries)``.

        """
        if betas is None:
            ll_temp = ll.copy()
        else:
            assert isinstance(betas, self.xp.ndarray)
            if ll.ndim > 1:
                betas_tmp = self.xp.expand_dims(betas, ll.ndim - 1)
            else:
                betas_tmp = betas
            ll_temp = betas_tmp * ll

        return ll_temp + lp

    def readout_adjustment(self, out_vals, all_vals_prop, aux_all_vals):
        """Read out values from the proposal.

        Allows the user to read out any values from the proposal that may be needed elsewhere. This function must be overwritten.

        Args:
            out_vals (list): ``[logP_out, ll_out, lp_out, log_proposal_pdf_out, log_sum_weights]``.
            all_vals_prop (list): ``[logP, ll, lp, log_proposal_pdf, log_sum_weights]``.
            aux_all_vals (list): ``[aux_logP, aux_ll, aux_lp, aux_log_proposal_pdf, aux_log_sum_weights]``.

        """
        pass

    def get_mt_proposal(
        self,
        coords,
        random,
        args_generate=(),
        kwargs_generate={},
        args_like=(),
        kwargs_like={},
        args_prior=(),
        kwargs_prior={},
        betas=None,
        ll_in=None,
        lp_in=None,
        inds_leaves_rj=None,
        inds_reverse_rj=None,
    ):
        """Make a multiple-try proposal

        Here, ``nwalkers`` refers to all independent walkers which generally
        will mean ``nwalkers * ntemps`` in terms of the rest of the sampler.

        Args:
            coords (np.ndarray): Current coordinates of walkers.
            random (obj): Random generator.
            args_generate (tuple, optional): Additional ``*args`` to pass to generate function.
                Must overwrite ``get_proposal`` function to use these.
                (default: ``()``)
            kwargs_generate (dict, optional): Additional ``**kwargs`` to pass to generate function.
                (default: ``{}``)
                Must overwrite ``get_proposal`` function to use these.
            args_like (tuple, optional): Additional ``*args`` to pass to Likelihood function.
                Must overwrite ``get_proposal`` function to use these.
                (default: ``()``)
            kwargs_like (dict, optional): Additional ``**kwargs`` to pass to Likelihood function.
                Must overwrite ``get_proposal`` function to use these.
                (default: ``{}``)
            args_prior (tuple, optional): Additional ``*args`` to pass to Prior function.
                Must overwrite ``get_proposal`` function to use these.
                (default: ``()``)
            kwargs_prior (dict, optional): Additional ``**kwargs`` to pass to Prior function.
                Must overwrite ``get_proposal`` function to use these.
                (default: ``{}``)
            betas (np.ndarray, optional): Inverse temperatures passes to the proposal with shape ``(nwalkers,)``.
            ll_in (np.ndarray, optional): Log Likelihood values coming in for current coordinates. Must be provided
                if ``self.rj is True``. If ``self.rj is True``, must be nested.
                Also, for all proposed removals, this value must be the Likelihood with the binary
                removed so all proposals are pretending to add a binary.
                Useful if ``self.independent is True``. (default: ``None``)
            lp_in (np.ndarray, optional): Log Prior values coming in for current coordinates. Must be provided
                if ``self.rj is True``. If ``self.rj is True``, must be nested.
                Also, for all proposed removals, this value must be the Likelihood with the binary
                removed so all proposals are pretending to add a binary.
                Useful if ``self.independent is True``. (default: ``None``)
            inds_leaves_rj (np.ndarray, optional): Array giving the leaf index of each incoming walker.
                Must be provided if ``self.rj is True``. (default: ``None``)
            inds_reverse_rj (np.ndarray, optional): Array giving the walker index for which proposals are
                reverse proposal removing a leaf.
                Must be provided if ``self.rj is True``. (default: ``None``)

        Returns:
            tuple: (generated points, factors).

        Raises:
            ValueError: Inputs are incorrect.

        """

        # check if rj and make sure we have all the information in that case
        if self.rj:
            try:
                assert ll_in is not None and lp_in is not None
                assert inds_leaves_rj is not None and inds_reverse_rj is not None
            except AssertionError:
                raise ValueError(
                    "If using rj, must provide ll_in, lp_in, inds_leaves_rj, and inds_reverse_rj."
                )

            # if using reversible jump, fill first spot with values that are proposed to remove
            fill_tuple = (inds_reverse_rj, np.zeros_like(inds_reverse_rj))
            fill_values = coords[inds_reverse_rj]
        else:
            fill_tuple = None
            fill_values = None

        # generate new points and get log of the proposal probability
        generated_points, log_proposal_pdf = self.special_generate_func(
            coords,
            random,
            *args_generate,
            size=self.num_try,
            fill_values=fill_values,
            fill_tuple=fill_tuple,
            **kwargs_generate
        )

        # compute the Likelihood functions
        ll = self.special_like_func(
            generated_points, *args_like, inds_leaves_rj=inds_leaves_rj, **kwargs_like
        )

        # check for nans
        if self.xp.any(self.xp.isnan(ll)):
            warnings.warn("Getting nans for ll in multiple try.")
            ll[self.xp.isnan(ll)] = -1e300

        # compute the Prior functions
        lp = self.special_prior_func(
            generated_points, *args_prior, inds_leaves_rj=inds_leaves_rj, **kwargs_prior
        )

        # if rj, make proposal distribution for all other leaves the prior value
        # this will properly cancel the prior with the proposal for leaves that already exists
        if self.rj:
            log_proposal_pdf += lp_in[:, None]

        # get posterior distribution including tempering
        logP = self.get_mt_log_posterior(ll, lp, betas=betas)

        log_importance_weights, log_sum_weights, inds_keep = get_mt_computations(
            logP, log_proposal_pdf, symmetric=self.symmetric, xp=self.xp
        )

        # tuple of index arrays of which try chosen per walker
        inds_tuple = (self.xp.arange(len(inds_keep)), inds_keep)

        if self.rj:
            # this just ensures the cancellation of logP and aux_logP outside of proposal
            inds_tuple[1][inds_reverse_rj] = 0

        # get chosen prior, Likelihood, posterior information
        lp_out = lp[inds_tuple]
        ll_out = ll[inds_tuple]
        logP_out = logP[inds_tuple]

        # store this information for access outside of multiple try part
        self.mt_lp = lp_out
        self.mt_ll = ll_out

        # choose points and get the log of the proposal for storage
        generated_points_out = generated_points[inds_tuple].copy()  # theta^j
        log_proposal_pdf_out = log_proposal_pdf[inds_tuple]

        # prepare auxillary information based on if it is nested rj, independent, or not
        if self.independent:
            # if independent, all the tries can be repeated for the auxillary draws
            aux_ll = ll.copy()
            aux_lp = lp.copy()

            # sub in the generation pdf for the current coordinates
            aux_log_proposal_pdf_sub = self.special_generate_logpdf(coords)

            # set sub ll based on if it is provided
            if ll_in is None:
                aux_ll_sub = self.special_generate_like(coords)

            else:
                assert ll_in.shape[0] == coords.shape[0]
                aux_ll_sub = ll_in

            # set sub lp based on if it is provided
            if lp_in is None:
                aux_lp_sub = self.special_generate_prior(coords)

            else:
                assert lp_in.shape[0] == coords.shape[0]
                aux_lp_sub = lp_in

            # sub in this information from the current coordinates
            aux_ll[inds_tuple] = aux_ll_sub
            aux_lp[inds_tuple] = aux_lp_sub

            # get auxillary posterior
            aux_logP = self.get_mt_log_posterior(aux_ll, aux_lp, betas=betas)

            # get aux_log_proposal_pdf information
            aux_log_proposal_pdf = log_proposal_pdf.copy()
            aux_log_proposal_pdf[inds_tuple] = aux_log_proposal_pdf_sub

            # set auxillary weights
            aux_log_importance_weights = aux_logP - aux_log_proposal_pdf

        elif self.rj:
            # in rj, set aux_ll and aux_lp to be repeats of the model with one less leaf
            aux_ll = np.repeat(ll_in[:, None], self.num_try, axis=-1)
            aux_lp = np.repeat(lp_in[:, None], self.num_try, axis=-1)

            # probability is the prior for the existing points
            aux_log_proposal_pdf = aux_lp.copy()

            # get log posterior
            aux_logP = self.get_mt_log_posterior(aux_ll, aux_lp, betas=betas)

            # get importance weights
            aux_log_importance_weights = aux_logP - aux_log_proposal_pdf

        else:
            # generate auxillary points based on chosen new points
            aux_generated_points, aux_log_proposal_pdf = self.special_generate_func(
                generated_points_out,
                random,
                *args_generate,
                size=self.num_try,
                fill_tuple=inds_tuple,
                fill_values=generated_points_out,
                **kwargs_generate
            )

            # get ll, lp, and lP
            aux_ll = self.special_like_func(
                aux_generated_points, *args_like, **kwargs_like
            )

            aux_lp = self.special_prior_func(aux_generated_points)

            aux_logP = self.get_mt_log_posterior(aux_ll, aux_lp, betas=betas)

            # set auxillary weights
            if not self.symmetric:
                aux_log_importance_weights = aux_logP - aux_log_proposal_pdf_sub
            else:
                aux_log_importance_weights = aux_logP

        # chosen output old Posteriors
        aux_logP_out = aux_logP[inds_tuple]
        # get sum of log weights
        aux_log_sum_weights = logsumexp(aux_log_importance_weights, axis=-1, xp=self.xp)

        aux_log_proposal_pdf_out = aux_log_proposal_pdf[inds_tuple]
        # this is setup to make clear with the math.
        # setting up factors properly means the
        # final lnpdiff will be effectively be the ratio of the sums
        # of the weights

        # IMPORTANT: logP_out must be subtracted against log_sum_weights before anything else due to -1e300s.
        factors = (
            (aux_logP_out - aux_log_sum_weights)
            - aux_log_proposal_pdf_out
            + aux_log_proposal_pdf_out
        ) - ((logP_out - log_sum_weights) - log_proposal_pdf_out + log_proposal_pdf_out)

        if self.rj:
            # adjust all information for reverese rj proposals
            factors[inds_reverse_rj] *= -1
            self.mt_ll[inds_reverse_rj] = ll_in[inds_reverse_rj]
            self.mt_lp[inds_reverse_rj] = lp_in[inds_reverse_rj]

        # store output information
        self.aux_logP_out = aux_logP_out
        self.logP_out = logP_out
        self.aux_ll = aux_ll
        self.aux_lp = aux_lp

        self.log_sum_weights = log_sum_weights
        self.aux_log_sum_weights = aux_log_sum_weights

        if self.rj:
            self.inds_reverse_rj = inds_reverse_rj
            self.inds_forward_rj = np.delete(
                np.arange(coords.shape[0]), inds_reverse_rj
            )

        # prepare to readout any information the user would like in readout_adjustment
        out_vals = [logP_out, ll_out, lp_out, log_proposal_pdf_out, log_sum_weights]
        all_vals_prop = [logP, ll, lp, log_proposal_pdf, log_sum_weights]
        aux_all_vals = [
            aux_logP,
            aux_ll,
            aux_lp,
            aux_log_proposal_pdf,
            aux_log_sum_weights,
        ]
        self.readout_adjustment(out_vals, all_vals_prop, aux_all_vals)

        return (
            generated_points_out,
            factors,
        )

    def get_proposal(self, branches_coords, random, branches_inds=None, **kwargs):
        """Get proposal

        Args:
            branches_coords (dict): Keys are ``branch_names`` and values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim] representing
                coordinates for walkers.
            random (object): Current random state object.
            branches_inds (dict, optional): Keys are ``branch_names`` and values are
                np.ndarray[ntemps, nwalkers, nleaves_max] representing which
                leaves are currently being used. (default: ``None``)
            **kwargs (ignored): This is added for compatibility. It is ignored in this function.

        Returns:
            tuple: (Proposed coordinates, factors) -> (dict, np.ndarray)

        Raises:
            ValueError: Input issues.

        """

        # mutliple try is only made for one branch here
        if len(list(branches_coords.keys())) > 1:
            raise ValueError("Can only propose change to one model at a time with MT.")

        # get main key
        key_in = list(branches_coords.keys())[0]
        self.key_in = key_in

        # get inds information
        if branches_inds is None:
            branches_inds = {}
            branches_inds[key_in] = np.ones(
                branches_coords[key_in].shape[:-1], dtype=bool
            )

        # Make sure for base proposals that there is only one leaf
        if np.any(branches_inds[key_in].sum(axis=-1) > 1):
            raise ValueError

        ntemps, nwalkers, _, _ = branches_coords[key_in].shape

        # get temperature information
        betas_here = np.repeat(
            self.temperature_control.betas[:, None],
            np.prod(branches_coords[key_in].shape[1:-1]),
        ).reshape(branches_inds[key_in].shape)[branches_inds[key_in]]

        # previous Likelihoods in case proposal is independent
        ll_here = np.repeat(
            self.current_state.log_like[:, :, None],
            branches_coords[key_in].shape[2],
            axis=-1,
        ).reshape(branches_inds[key_in].shape)[branches_inds[key_in]]

        # previous Priors in case proposal is independent
        lp_here = np.repeat(
            self.current_state.log_prior[:, :, None],
            branches_coords[key_in].shape[2],
            axis=-1,
        ).reshape(branches_inds[key_in].shape)[branches_inds[key_in]]

        # get mt proposal
        generated_points, factors = self.get_mt_proposal(
            branches_coords[key_in][branches_inds[key_in]],
            random,
            betas=betas_here,
            ll_in=ll_here,
            lp_in=lp_here,
        )

        # store this information for access outside
        self.mt_ll = self.mt_ll.reshape(ntemps, nwalkers)
        self.mt_lp = self.mt_lp.reshape(ntemps, nwalkers)

        return (
            {key_in: generated_points.reshape(ntemps, nwalkers, 1, -1)},
            factors.reshape(ntemps, nwalkers),
        )


class MultipleTryMoveRJ(MultipleTryMove):
    def get_proposal(
        self,
        branches_coords,
        branches_inds,
        nleaves_min_all,
        nleaves_max_all,
        random,
        **kwargs
    ):
        """Make a proposal

        Args:
            all_coords (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim]. These are the curent
                coordinates for all the walkers.
            all_inds (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max]. These are the boolean
                arrays marking which leaves are currently used within each walker.
            nleaves_min_all (list): Minimum values of leaf ount for each model. Must have same order as ``all_cords``.
            nleaves_max_all (list): Maximum values of leaf ount for each model. Must have same order as ``all_cords``.
            random (object): Current random state of the sampler.
            **kwargs (ignored): For modularity.

        Returns:
            tuple: Tuple containing proposal information.
                First entry is the new coordinates as a dictionary with keys
                as ``branch_names`` and values as
                ``double `` np.ndarray[ntemps, nwalkers, nleaves_max, ndim] containing
                proposed coordinates. Second entry is the new ``inds`` array with
                boolean values flipped for added or removed sources. Third entry
                is the factors associated with the
                proposal necessary for detailed balance. This is effectively
                any term in the detailed balance fraction. +log of factors if
                in the numerator. -log of factors if in the denominator.

        """

        if len(list(branches_coords.keys())) > 1:
            raise ValueError("Can only propose change to one model at a time with MT.")

        # get main key
        key_in = list(branches_coords.keys())[0]
        self.key_in = key_in

        if branches_inds is None:
            raise ValueError("In MT RJ proposal, branches_inds cannot be None.")

        ntemps, nwalkers, nleaves_max, ndim = branches_coords[key_in].shape

        # get temperature information
        betas_here = np.repeat(
            self.temperature_control.betas[:, None], nwalkers, axis=-1
        ).flatten()

        # current Likelihood and prior information
        ll_here = self.current_state.log_like.flatten()
        lp_here = self.current_state.log_prior.flatten()

        # do rj setup
        assert len(nleaves_min_all) == 1 and len(nleaves_max_all) == 1
        nleaves_min = nleaves_min_all[key_in]
        nleaves_max = nleaves_max_all[key_in]

        if nleaves_min == nleaves_max:
            raise ValueError("MT RJ proposal requires that nleaves_min != nleaves_max.")
        elif nleaves_min > nleaves_max:
            raise ValueError("nleaves_min is greater than nleaves_max. Not allowed.")

        # get the inds adjustment information
        all_inds_for_change = self.get_model_change_proposal(
            branches_inds[key_in], random, nleaves_min, nleaves_max
        )

        # preparing leaf information for going into the proposal
        inds_leaves_rj = np.zeros(ntemps * nwalkers, dtype=int)
        coords_in = np.zeros((ntemps * nwalkers, ndim))
        inds_reverse_rj = None

        # prepare proposal dictionaries
        new_inds = deepcopy(branches_inds)
        q = deepcopy(branches_coords)
        for change in all_inds_for_change.keys():
            if change not in ["+1", "-1"]:
                raise ValueError("MT RJ is only implemented for +1/-1 moves.")

            # get indicies of changing leaves
            temp_inds = all_inds_for_change[change][:, 0]
            walker_inds = all_inds_for_change[change][:, 1]
            leaf_inds = all_inds_for_change[change][:, 2]

            # leaf index to change
            inds_leaves_rj[temp_inds * nwalkers + walker_inds] = leaf_inds
            coords_in[temp_inds * nwalkers + walker_inds] = branches_coords[key_in][
                (temp_inds, walker_inds, leaf_inds)
            ]

            # adjustment of indices
            new_val = {"+1": True, "-1": False}[change]

            # adjust indices
            new_inds[key_in][(temp_inds, walker_inds, leaf_inds)] = new_val

            if change == "-1":
                # which walkers are removing
                inds_reverse_rj = temp_inds * nwalkers + walker_inds

        if inds_reverse_rj is not None:
            # setup reversal coords and inds
            # need to determine Likelihood and prior of removed binaries.
            # this goes into the multiple try proposal as previous ll and lp
            temp_reverse_coords = {}
            temp_reverse_inds = {}

            for key in self.current_state.branches:
                (
                    ntemps_tmp,
                    nwalkers_tmp,
                    nleaves_max_tmp,
                    ndim_tmp,
                ) = self.current_state.branches[key].shape

                # coords from reversal
                temp_reverse_coords[key] = self.current_state.branches[
                    key
                ].coords.reshape(ntemps_tmp * nwalkers_tmp, nleaves_max_tmp, ndim_tmp)[
                    inds_reverse_rj
                ][
                    None, :
                ]

                # which inds array to use
                inds_tmp_here = (
                    new_inds[key]
                    if key == key_in
                    else self.current_state.branches[key].inds
                )
                temp_reverse_inds[key] = inds_tmp_here.reshape(
                    ntemps * nwalkers, nleaves_max_tmp
                )[inds_reverse_rj][None, :]

            # calculate information for the reverse
            lp_reverse_here = self.current_model.compute_log_prior_fn(
                temp_reverse_coords, inds=temp_reverse_inds
            )[0]
            ll_reverse_here = self.current_model.compute_log_like_fn(
                temp_reverse_coords, inds=temp_reverse_inds, logp=lp_here
            )[0]

            # fill the here values
            ll_here[inds_reverse_rj] = ll_reverse_here
            lp_here[inds_reverse_rj] = lp_reverse_here

        # get mt proposal
        generated_points, factors = self.get_mt_proposal(
            coords_in,
            random,
            betas=betas_here,
            ll_in=ll_here,
            lp_in=lp_here,
            inds_leaves_rj=inds_leaves_rj,
            inds_reverse_rj=inds_reverse_rj,
        )

        # for reading outside
        self.mt_ll = self.mt_ll.reshape(ntemps, nwalkers)
        self.mt_lp = self.mt_lp.reshape(ntemps, nwalkers)

        # which walkers have information added
        inds_forward_rj = np.delete(np.arange(coords_in.shape[0]), inds_reverse_rj)

        # updated the coordinates
        temp_inds = all_inds_for_change["+1"][:, 0]
        walker_inds = all_inds_for_change["+1"][:, 1]
        leaf_inds = all_inds_for_change["+1"][:, 2]
        q[key_in][(temp_inds, walker_inds, leaf_inds)] = generated_points[
            inds_forward_rj
        ]

        return q, new_inds, factors.reshape(ntemps, nwalkers)



# -*- coding: utf-8 -*-

from collections import namedtuple

__all__ = ["Model"]


Model = namedtuple(
    "Model",
    (
        "log_like_fn",
        "compute_log_like_fn",
        "compute_log_prior_fn",
        "temperature_control",
        "map_fn",
        "random",
    ),
)



# -*- coding: utf-8 -*-

import numpy as np

from .mh import MHMove

__all__ = ["GaussianMove"]


class GaussianMove(MHMove):
    """A Metropolis step with a Gaussian proposal function.

    This class is heavily based on the same class in ``emcee``.

    Args:
        cov (dict): The covariance of the proposal function. The keys are branch names and the
            values are covariance information. This information can be provided as a scalar,
            vector, or matrix and the proposal will be assumed isotropic,
            axis-aligned, or general, respectively.
        mode (str, optional): Select the method used for updating parameters. This
            can be one of ``"vector"``, ``"random"``, or ``"sequential"``. The
            ``"vector"`` mode updates all dimensions simultaneously,
            ``"random"`` randomly selects a dimension and only updates that
            one, and ``"sequential"`` loops over dimensions and updates each
            one in turn. (default: ``"vector"``)
        factor (float, optional): If provided the proposal will be made with a
            standard deviation uniformly selected from the range
            ``exp(U(-log(factor), log(factor))) * cov``. This is invalid for
            the ``"vector"`` mode. (default: ``None``)
        **kwargs (dict, optional): Kwargs for parent classes. (default: ``{}``)

    Raises:
        ValueError: If the proposal dimensions are invalid or if any of any of
            the other arguments are inconsistent.

    """

    def __init__(self, cov_all, mode="vector", factor=None, **kwargs):
        self.all_proposal = {}
        for name, cov in cov_all.items():
            # Parse the proposal type.
            try:
                float(cov)

            except TypeError:
                cov = np.atleast_1d(cov)
                if len(cov.shape) == 1:
                    # A diagonal proposal was given.
                    ndim = len(cov)
                    proposal = _diagonal_proposal(np.sqrt(cov), factor, mode)

                elif len(cov.shape) == 2 and cov.shape[0] == cov.shape[1]:
                    # The full, square covariance matrix was given.
                    ndim = cov.shape[0]
                    proposal = _proposal(cov, factor, mode)

                else:
                    raise ValueError("Invalid proposal scale dimensions")

            else:
                # This was a scalar proposal.
                ndim = None
                proposal = _isotropic_proposal(np.sqrt(cov), factor, mode)
            self.all_proposal[name] = proposal

        super(GaussianMove, self).__init__(**kwargs)

    def get_proposal(self, branches_coords, random, branches_inds=None, **kwargs):
        """Get proposal from Gaussian distribution

        Args:
            branches_coords (dict): Keys are ``branch_names`` and values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim] representing
                coordinates for walkers.
            random (object): Current random state object.
            branches_inds (dict, optional): Keys are ``branch_names`` and values are
                np.ndarray[ntemps, nwalkers, nleaves_max] representing which
                leaves are currently being used. (default: ``None``)
            **kwargs (ignored): This is added for compatibility. It is ignored in this function.

        Returns:
            tuple: (Proposed coordinates, factors) -> (dict, np.ndarray)

        """

        # initialize ouput
        q = {}
        for name, coords in zip(branches_coords.keys(), branches_coords.values()):
            ntemps, nwalkers, nleaves_max, ndim = coords.shape

            # setup inds accordingly
            if branches_inds is None:
                inds = np.ones((ntemps, nwalkers, nleaves_max), dtype=bool)
            else:
                inds = branches_inds[name]

            # get the proposal for this branch
            proposal_fn = self.all_proposal[name]
            inds_here = np.where(inds == True)

            # copy coords
            q[name] = coords.copy()

            # get new points
            new_coords, _ = proposal_fn(coords[inds_here], random)

            # put into coords in proper location
            q[name][inds_here] = new_coords.copy()

        # handle periodic parameters
        if self.periodic is not None:
            q = self.periodic.wrap(
                {
                    name: tmp.reshape((ntemps * nwalkers,) + tmp.shape[-2:])
                    for name, tmp in q.items()
                },
                xp=self.xp,
            )

            q = {
                name: tmp.reshape(
                    (
                        ntemps,
                        nwalkers,
                    )
                    + tmp.shape[-2:]
                )
                for name, tmp in q.items()
            }

        return q, np.zeros((ntemps, nwalkers))


class _isotropic_proposal(object):
    allowed_modes = ["vector", "random", "sequential"]

    def __init__(self, scale, factor, mode):
        self.index = 0
        self.scale = scale
        
        if isinstance(scale, float):
            self.invscale = 1. / scale
        else:
            self.invscale = np.linalg.inv(np.linalg.cholesky(scale))

        if factor is None:
            self._log_factor = None
        else:
            if factor < 1.0:
                raise ValueError("'factor' must be >= 1.0")
            self._log_factor = np.log(factor)

        if mode not in self.allowed_modes:
            raise ValueError(
                ("'{0}' is not a recognized mode. " "Please select from: {1}").format(
                    mode, self.allowed_modes
                )
            )
        self.mode = mode

    def get_factor(self, rng):
        if self._log_factor is None:
            return 1.0
        return np.exp(rng.uniform(-self._log_factor, self._log_factor))

    def get_updated_vector(self, rng, x0):
        return x0 + self.get_factor(rng) * self.scale * rng.randn(*(x0.shape))

    def __call__(self, x0, rng):
        nw, nd = x0.shape
        xnew = self.get_updated_vector(rng, x0)
        if self.mode == "random":
            m = (range(nw), rng.randint(x0.shape[-1], size=nw))
        elif self.mode == "sequential":
            m = (range(nw), self.index % nd + np.zeros(nw, dtype=int))
            self.index = (self.index + 1) % nd
        else:
            return xnew, np.zeros(nw)
        x = np.array(x0)
        x[m] = xnew[m]
        return x, np.zeros(nw)


class _diagonal_proposal(_isotropic_proposal):
    def get_updated_vector(self, rng, x0):
        return x0 + self.get_factor(rng) * self.scale * rng.randn(*(x0.shape))


class _proposal(_isotropic_proposal):
    allowed_modes = ["vector"]

    def get_updated_vector(self, rng, x0):
        return x0 + self.get_factor(rng) * rng.multivariate_normal(
            np.zeros(len(self.scale)), self.scale, size=len(x0)
        )



# -*- coding: utf-8 -*-
try:
    import cupy as cp
except (ModuleNotFoundError, ImportError):
    pass

import numpy as np

from .red_blue import RedBlueMove

__all__ = ["StretchMove"]


class StretchMove(RedBlueMove):
    """Affine-Invariant Proposal

    A `Goodman & Weare (2010)
    <https://msp.org/camcos/2010/5-1/p04.xhtml>`_ "stretch move" with
    parallelization as described in `Foreman-Mackey et al. (2013)
    <https://arxiv.org/abs/1202.3665>`_.

    This class was originally implemented in ``emcee``.

    Args:
        a (double, optional): The stretch scale parameter. (default: ``2.0``)
        return_gpu (bool, optional): If ``use_gpu == True and return_gpu == True``,
            the returned arrays will be returned as ``CuPy`` arrays. (default: ``False``)
        kwargs (dict, optional): Additional keyword arguments passed down through :class:`RedRedBlueMove`_.

    Attributes:
        a (double): The stretch scale parameter.
        return_gpu (bool): Whether the array being returned is in ``Cupy`` (``True``)
            or ``NumPy`` (``False``).

    """

    def __init__(self, a=2.0, return_gpu=False, random_seed=None, **kwargs):
        # store scale factor
        self.a = a

        # pass kwargs up
        RedBlueMove.__init__(self, **kwargs)

        # change array library based on GPU usage

        # set the random seet of the library if desired
        if random_seed is not None:
            self.xp.random.seed(random_seed)

        self.return_gpu = return_gpu

        # how it was formerly
        # super(StretchMove, self).__init__(**kwargs)

    def adjust_factors(self, factors, ndims_old, ndims_new):
        """Adjust the ``factors`` based on changing dimensions.

        ``factors`` is adjusted in place.

        Args:
            factors (xp.ndarray): Array of ``factors`` values. It is adjusted in place.
            ndims_old (int or xp.ndarray): Old dimension. If given as an ``xp.ndarray``,
                must be broadcastable with ``factors``.
            ndims_new (int or xp.ndarray): New dimension. If given as an ``xp.ndarray``,
                must be broadcastable with ``factors``.

        """
        # adjusts in place
        if ndims_old == ndims_new:
            return
        logzz = factors / (ndims_old - 1.0)
        factors[:] = logzz * (ndims_new - 1.0)

    def choose_c_vals(self, c, Nc, Ns, ntemps, random_number_generator, **kwargs):
        """Get the compliment array

        The compliment represents the points that are used to move the actual points whose position is
        changing.

        Args:
            c (np.ndarray): Possible compliment values with shape ``(ntemps, Nc, nleaves_max, ndim)``.
            Nc (int): Length of the ``...``: the subset of walkers proposed to move now (usually nwalkers/2).
            Ns (int): Number of generation points.
            ntemps (int): Number of temperatures.
            random_number_generator (object): Random state object.
            **kwargs (ignored): Ignored here. For modularity.

        Returns:
            np.ndarray: Compliment values to use with shape ``(ntemps, Ns, nleaves_max, ndim)``.

        """

        rint = random_number_generator.randint(
            Nc,
            size=(
                ntemps,
                Ns,
            ),
        )
        c_temp = self.xp.take_along_axis(c, rint[:, :, None, None], axis=1)
        return c_temp

    def get_new_points(
        self, name, s, c_temp, Ns, branch_shape, branch_i, random_number_generator
    ):
        """Get mew points in stretch move.

        Takes compliment and uses it to get new points for those being proposed.

        Args:
            name (str): Branch name.
            s (np.ndarray): Points to be moved with shape ``(ntemps, Ns, nleaves_max, ndim)``.
            c (np.ndarray): Compliment to move points with shape ``(ntemps, Ns, nleaves_max, ndim)``.
            Ns (int): Number to generate.
            branch_shape (tuple): Full branch shape.
            branch_i (int): Which branch in the order is being run now. This ensures that the
                randomly generated quantity per walker remains the same over branches.
            random_number_generator (object): Random state object.

        Returns:
            np.ndarray: New proposed points with shape ``(ntemps, Ns, nleaves_max, ndim)``.


        """

        ntemps, nwalkers, nleaves_max, ndim_here = branch_shape

        # only for the first branch do we draw for zz
        if branch_i == 0:
            self.zz = (
                (self.a - 1.0) * random_number_generator.rand(ntemps, Ns) + 1
            ) ** 2.0 / self.a

        # get proper distance

        if self.periodic is not None:
            diff = self.periodic.distance(
                {name: s.reshape(ntemps * nwalkers, nleaves_max, ndim_here)},
                {name: c_temp.reshape(ntemps * nwalkers, nleaves_max, ndim_here)},
                xp=self.xp,
            )[name].reshape(ntemps, nwalkers, nleaves_max, ndim_here)
        else:
            diff = c_temp - s

        temp = c_temp - (diff) * self.zz[:, :, None, None]

        # wrap periodic values

        if self.periodic is not None:
            temp = self.periodic.wrap(
                {name: temp.reshape(ntemps * nwalkers, nleaves_max, ndim_here)},
                xp=self.xp,
            )[name].reshape(ntemps, nwalkers, nleaves_max, ndim_here)

        # get from gpu or not
        if self.use_gpu and not self.return_gpu:
            temp = temp.get()
        return temp

    def get_proposal(self, s_all, c_all, random, gibbs_ndim=None, **kwargs):
        """Generate stretch proposal

        Args:
            s_all (dict): Keys are ``branch_names`` and values are coordinates
                for which a proposal is to be generated.
            c_all (dict): Keys are ``branch_names`` and values are lists. These
                lists contain all the complement array values.
            random (object): Random state object.
            gibbs_ndim (int or np.ndarray, optional): If Gibbs sampling, this indicates
                the true dimension. If given as an array, must have shape ``(ntemps, nwalkers)``.
                See the tutorial for more information.
                (default: ``None``)

        Returns:
            tuple: First entry is new positions. Second entry is detailed balance factors.

        Raises:
            ValueError: Issues with dimensionality.

        """

        # needs to be set before we reach the end
        self.zz = None
        random_number_generator = random if not self.use_gpu else self.xp.random
        newpos = {}

        # iterate over branches
        for i, name in enumerate(s_all):
            # get points to move
            s = self.xp.asarray(s_all[name])

            if not isinstance(c_all[name], list):
                raise ValueError("c_all for each branch needs to be a list.")

            # get compliment possibilities
            c = [self.xp.asarray(c_tmp) for c_tmp in c_all[name]]

            ntemps, nwalkers, nleaves_max, ndim_here = s.shape
            c = self.xp.concatenate(c, axis=1)

            Ns, Nc = s.shape[1], c.shape[1]
            # gets rid of any values of exactly zero
            ndim_temp = nleaves_max * ndim_here

            # need to properly handle ndim
            if i == 0:
                ndim = ndim_temp
                Ns_check = Ns

            else:
                ndim += ndim_temp
                if Ns_check != Ns:
                    raise ValueError("Different number of walkers across models.")

            # get actual compliment values
            c_temp = self.choose_c_vals(c, Nc, Ns, ntemps, random_number_generator)

            # use stretch to get new proposals
            newpos[name] = self.get_new_points(
                name, s, c_temp, Ns, s.shape, i, random_number_generator
            )
        # proper factors
        factors = (ndim - 1.0) * self.xp.log(self.zz)
        if self.use_gpu and not self.return_gpu:
            factors = factors.get()

        if gibbs_ndim is not None:
            # adjust factors in place
            self.adjust_factors(factors, ndim, gibbs_ndim)

        return newpos, factors
